/* arslds.c
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 */

#include <math.h>
#include <stdlib.h>

#include "arslds.h"
#include "linalg.h"
#include "utils.h"
/*-------------------------------------------------------------------------*/
void arslds_fwd_sgl(arslds_t *arslds, int T, double *v, 
		    double **gainH, double **gainV, double **gainN,
		    double **logr, darray *f, darray *F, double *llh) {

  int i, n;
  int st, stm, t, cnt;

  double logpV1;
  double logpvtgV1tm;

  double half_log2pi = 0.5*log(2*M_PI);

  /* Get ARSLDS parameters */

  double *prior  = arslds->prior;
  double **stran = arslds->stran;

  matrix **pMean = arslds->pMean;
  matrix **A     = arslds->A;
  double *CovV   = arslds->CovV;
  double *vbias  = arslds->vbias;

  int S = arslds->S;
  int H = arslds->H;
  int K = arslds->K;
  int D = arslds->D;

  int E = H - D;

  /* To be allocated */

  int *perm;

  double **logstran;
  double **logpstmstgV1t;

  matrix *tmp[2];

  darray *meanXgY;
  darray *SigmaXgY;

  /* Allocate memory */

  perm = (int*)malloc(sizeof(int));

  logstran = (double**)malloc(S*sizeof(double*));

  for (stm=0; stm<S; ++stm)
    logstran[stm] = (double*)malloc(S*sizeof(double));

  logpstmstgV1t = (double**)malloc(S*sizeof(double*));

  for (st=0; st<S; ++st)
    logpstmstgV1t[st] = (double*)malloc(S*sizeof(double));

  meanXgY  = darray_alloc(3, H, S, S);
  SigmaXgY = darray_alloc(4, H, H, S, S);

  /* First time step */

  tmp[0] = matrix_alloc(H,1);

  cnt = 0;
  n   = 0;

  for (st=0; st<S; ++st) {
    /* Is it necessary to compute everything? */

    if (prior[st] == 0.0) {
      logr[0][st] = -INFINITY;
      
      continue;
    } else {
      
      matrix *mht, *DhtDht, *K;
      double mvt, DvtDvt, z;

      matrix ft = darray_getmat(f, 1, st, 0); 
      matrix Ft = darray_getmat(F, 2, st, 0);
      
      /* f(:,st,1) = pMean(:,st); mht = f(:,st,1) */
      
      matrix_memcpy(&ft, pMean[st]);
      mht = &ft;
      
      /* F(:,:,st,1) = gainH(:,:,st,n); Dhtht = F(:,:,st,1) */
    
      matrix_init(&Ft, 0);
      matrix_set(&Ft, 0, 0, gainH[n][st]);

      if (gainN) {
	/* F(:,:,st,1) = F(:,:,st,1) + gainN(:,:,st,n) */

	matrix_set(&Ft, E, E, gainN[n][st]); 
      }

      DhtDht = &Ft;

      /* mvt = B(:,:,st)*mht + vbias(:,st) */
      
      mvt = matrix_get(mht, 0, 0) + vbias[st];

      if (gainN) {
	mvt += matrix_get(mht, E, 0);
      }

      /* DvtDvt = B(:,:,st)*DhtDht*B(:,:,st)' */

      DvtDvt = matrix_get(DhtDht, 0, 0);

      if (gainV) {
	/* DvtDvt = DvtDvt + gainV(:,:,st,n) */
    
	DvtDvt += gainV[n][st];
      } else {
	/* DvtDvt = DvtDvt + CovV(:,:,st) */
    
	DvtDvt += CovV[st];
      }

      if (gainN)
	DvtDvt += gainN[n][st];
	
      /* K = DhtDht*B(:,:,st) */

      {
	matrix m = matrix_submatrix(DhtDht, 0, H-1, 0, 0);
	matrix_memcpy(tmp[0], &m);

	if (gainN) {
	  matrix m = matrix_submatrix(DhtDht, 0, H-1, E, E);
	  
	  matrix_add(&m, tmp[0], tmp[0]);
	}

	K = tmp[0];
      }

      /* F(:,:,st,1) = F(:,:,st,1) - K*K' / DvtDvt */

      matrix_dgemm(NoTrans, Trans, -1.0/DvtDvt, K, K, 1.0, &Ft);

      /* z = v(:,1) - mvt */
    
      z = v[0] - mvt;

      /* f(:,st,1) = f(:,st,1) + K*z */

      matrix_addn(NoTrans, NoTrans, 1.0, &ft, z/DvtDvt, K, &ft);

      /* logr(st,1) = log(prior(st)) - 0.5 * log(DvtDvt) 
	 - 0.5 * log(2*pi) */
      
      logr[0][st] = log(prior[st]) - 0.5 * log(DvtDvt) - half_log2pi;
      
      /* logr(st,1) = logr(st,1) - 0.5 * z^2 / DvtDvt */

      logr[0][st] -= 0.5 * z*z / DvtDvt;
    }
  }

  /* Normalise the messages and update the llh */
  
  logpV1 = logsum(S, logr[0]);
  
  for (i=0; i<S; ++i)
    logr[0][i] -= logpV1;

  *llh = logpV1;

  /* Deallocate temporary variables */

  matrix_free(tmp[0]);

  /* Remaining time steps */

  tmp[0] = matrix_alloc(H,H);
  tmp[1] = matrix_alloc(H,1);

  /* ... */

  for (stm=0; stm<S; ++stm)
    for (st=0; st<S; ++st)
      logstran[stm][st] = log(stran[stm][st]);
  
  /* ... */
  
  for (t=1; t<T; ++t) {
    int m = (t-1) / K;

    cnt = (cnt-1+K) % K;
    
    if (cnt == 0) ++n;
    
    for (st=0; st<S; ++st) {
      for (stm=0; stm<S; ++stm) {
	/* Is it necessary to compute everything? */
	
	if ((cnt != 0 && st != stm) ||
	    logstran[stm][st] == -INFINITY || 
	    logr[t-1][stm] == -INFINITY) {

	  logpstmstgV1t[st][stm] = -INFINITY;
	  
	  continue;
	} else {
	  double logpvtgststmV1tm, a, mvt, z, DvtDvt;
	  
	  matrix *DhtDht, *K;
	  
	  matrix ftm = darray_getmat(f, 1, stm, t-1);
	  matrix Ftm = darray_getmat(F, 2, stm, m);
	  
	  matrix mhtgV1t    = darray_getmat(meanXgY, 1, stm, st);
	  matrix DhtDhtgV1t = darray_getmat(SigmaXgY, 2, stm, st);
	  
	  /* mhtgV1t = A(:,:,st)*ftm */
	  
	  matrix_dgemm(NoTrans, NoTrans, 1.0, A[st], &ftm, 
		       0.0, &mhtgV1t);

	  /* mvt = mhtgV1t(1) + vbias(:,st) */
	  
	  mvt = matrix_get(&mhtgV1t, 0, 0) + vbias[st];

	  if (gainN) {
	    mvt += matrix_get(&mhtgV1t, E, 0);
	  }
	  
	  /* DhtDhtgV1t = A(:,:,st)*Ftm*A(:,:,st)' + gainH(:,:,st,n) */
	  
	  matrix_dgemm(NoTrans, Trans, 1.0, &Ftm, A[st], 0.0, tmp[0]);
	  matrix_dgemm(NoTrans, NoTrans, 1.0, A[st], tmp[0], 
		       0.0, &DhtDhtgV1t);
	  
	  a = matrix_get(&DhtDhtgV1t, 0, 0);
	  matrix_set(&DhtDhtgV1t, 0, 0, a + gainH[n][st]);

	  if (gainN) {
	    double b = matrix_get(&DhtDhtgV1t, E, E);
	    matrix_set(&DhtDhtgV1t, E, E, b + gainN[n][st]);
	  }
	  
	  /* DhtDht = DhtDhtgV1t */
	  
	  DhtDht = &DhtDhtgV1t;
	  
	  /* DvtDvt = B(:,:,st)*DhtDht*B(:,:,st)' */

	  DvtDvt = a + gainH[n][st];

	  if (gainV) {
	    /* DvtDvt = DvtDvt + gainV(:,:,st,n) */
	    
	    DvtDvt += gainV[n][st];
	  } else {
	    /* DvtDvt = DvtDvt + CovV(:,:,st) */
	    
	    DvtDvt += CovV[st];
	  }

	  if (gainN)
	    DvtDvt += gainN[n][st];
	  
	  /* K = DhtDht*B(:,:,st) */
	  
	  {
	    matrix m = matrix_submatrix(DhtDht, 0, H-1, 0, 0);
	    matrix_memcpy(tmp[1], &m);

	    if (gainN) {
	      matrix m = matrix_submatrix(DhtDht, 0, H-1, E, E);
	      
	      matrix_add(&m, tmp[1], tmp[1]);
	    }
	    
	    K = tmp[1];
	  }
	  
	  /* DhtDhtgV1t = DhtDhtgV1t - K*K' / DvtDvt */
	  
	  matrix_dgemm(NoTrans, Trans, -1.0/DvtDvt, K, K, 1.0, &DhtDhtgV1t);
	  
	  /* z = v(:,t) - mvt */
	  
	  z = v[t] - mvt;
	  
	  /* mhtgV1t = mhtgV1t + K*z */
	  
	  matrix_addn(NoTrans, NoTrans, 1.0, &mhtgV1t, 
		      z/DvtDvt, K, &mhtgV1t);
	  
	  /* logpvtgststmV1tm = - 0.5 * log(DvtDvt) - 0.5 * log(2*pi) */ 
	  
	  logpvtgststmV1tm = -0.5 * log(DvtDvt) - half_log2pi;
	  
	  /* logpvtgststmV1tm = logpvtgststmV1tm 
	     - 0.5 * z^2 / DvtDvt */
	  
	  logpvtgststmV1tm -= 0.5 * z*z / DvtDvt;
	  
	  /* logpstmstgV1t(stm,st) = logpvtgststmV1tm
	     + logstran(st, stm)
	     + logr(stm, t-1) */
	  
	  logpstmstgV1t[st][stm] = 
	    logpvtgststmV1tm
	    + (cnt == 0 ? logstran[stm][st] : 0.0)
	    + logr[t-1][stm];
	}
      }
    }
    
    /* Update the llh */
    
    logpvtgV1tm = -INFINITY;
    
    for (st=0; st<S; ++st)
      for (stm=0; stm<S; ++stm)
	logpvtgV1tm = logadd(logpvtgV1tm, logpstmstgV1t[st][stm]);

    *llh = *llh + logpvtgV1tm;
    
    /* logpstmstgV1t = logpstmstgV1t - logpvtgV1tm */
    
    for (st=0; st<S; ++st)
      for (stm=0; stm<S; ++stm)
	logpstmstgV1t[st][stm] -= logpvtgV1tm;
    
    /* collapse */
    
    for (st=0; st<S; ++st) {
      /* r(st,t) = p(st | v_1:t) */
      
      double logpstgV1t = -INFINITY;
	
      for (stm=0; stm<S; ++stm)
	logpstgV1t = logadd(logpstgV1t, logpstmstgV1t[st][stm]);
      
      logr[t][st] = logpstgV1t;
      
      if (logpstgV1t != -INFINITY) {
	
	matrix ft = darray_getmat(f, 1, st, t);
	matrix Ft = darray_getmat(F, 2, st, n);
	
	darray means = darray_subarray(meanXgY, 2, st);
	darray Covs  = darray_subarray(SigmaXgY, 3, st);

	matrix_init(&Ft, 0.0);

	arslds_mix2sgl(S, logpstmstgV1t[st], logpstgV1t, &means, 
		       &Covs, &ft, &Ft);
      }
    }
  }
  
  *llh /= T;
  
  /* Deallocate temporary variables */
  
  matrix_free(tmp[0]);
  matrix_free(tmp[1]);
  
  /* Deallocate memory */
  
  darray_free(SigmaXgY);
  darray_free(meanXgY);  
  
  for (i=0; i<S; ++i) {
    free(logstran[i]);
    free(logpstmstgV1t[i]);
  }
  
  free(logstran);
  free(logpstmstgV1t);
  
  free(perm);
}
/*-------------------------------------------------------------------------*/
void arslds_bwd_sgl(arslds_t *arslds, int T, 
		    double **gainH, double **gainN,
		    double **logr, darray *f, darray *F, double **logl, 
		    darray *pair_mean, darray *pair_2mom, int useKim) {
  
  int i;
  int st, stp, t, cnt;

  double half_log2pi = 0.5*log(2*M_PI);

  /* Get ARSLDS parameters */

  int S = arslds->S;
  int H = arslds->H;
  int K = arslds->K;
  int D = arslds->D;
  int N = T / K + (T % K == 0 ? 0 : 1);

  int E = H - D;

  double **stran = arslds->stran;

  matrix **A = arslds->A;

  double avtmp = 0.0;

  /* To be allocated */

  int *perm;
  
  double **logstran;

  double **logpstgstpV1THtp;
  double **logpstpstgV1T;

  matrix *tmp[9];

  darray *mHtV1T;
  darray *CovHtV1T;
  darray *CovHtHtpV1T;

  darray *CovHtpHtgV1t;
  darray *CovHtpHtpgV1t;
  darray *leftAp;

  /* Allocate memory */

  perm = (int*)malloc(H*sizeof(int));

  logstran = (double**)malloc(S*sizeof(double*));

  for (i=0; i<S; ++i)
    logstran[i] = (double*)malloc(S*sizeof(double));

  logpstgstpV1THtp = (double**)malloc(S*sizeof(double*));
  logpstpstgV1T    = (double**)malloc(S*sizeof(double*));

  for (st=0; st<S; ++st) {
    logpstgstpV1THtp[st] = (double*)malloc(S*sizeof(double));
    logpstpstgV1T[st]    = (double*)malloc(S*sizeof(double));
  }

  mHtV1T      = darray_alloc(3, H, S, S);
  CovHtV1T    = darray_alloc(4, H, H, S, S);
  CovHtHtpV1T = darray_alloc(4, H, H, S, S);

  CovHtpHtgV1t  = darray_alloc(4, H, H, S, S);
  CovHtpHtpgV1t = darray_alloc(4, H, H, S, S);
  leftAp        = darray_alloc(4, H, H, S, S);

  /* Get the log of the discrete transition matrix */

  for (st=0; st<S; ++st)
    for (stp=0; stp<S; ++stp)
      logstran[st][stp] = log(stran[st][stp]);

  /* Message initialisation */

  for (st=0; st<S; ++st) {
    matrix pmean = darray_getmat(pair_mean, 1, st, T-1);
    matrix p2mom = darray_getmat(pair_2mom, 2, st, T-1);

    matrix gt = matrix_submatrix(&pmean, H, 2*H-1, 0, 0);
    matrix Gt = matrix_submatrix(&p2mom, H, 2*H-1, H, 2*H-1);

    matrix ft = darray_getmat(f, 1, st, T-1);
    matrix Ft = darray_getmat(F, 2, st, N-1);
    
    matrix_memcpy(&gt, &ft);
    matrix_memcpy(&Gt, &Ft);  /* This will be converted to 2mom later */

    logl[T-1][st] = logr[T-1][st];
  }
  
  /* ... */

  tmp[0]  = matrix_alloc(H,1);
  tmp[1]  = matrix_alloc(H,H);
  tmp[2]  = matrix_alloc(H,H);
  tmp[3]  = matrix_alloc(H,H);
  tmp[4]  = matrix_alloc(H,H);
  tmp[5]  = matrix_alloc(H,1);
  tmp[6]  = matrix_alloc(H,H);
  tmp[7]  = matrix_alloc(H,1);
  tmp[8]  = matrix_alloc(H,H);

  {
    int n = 1 + T / K;
    cnt = (n*K - T + 1) % K;
  }
  
  for (t=T-2; t>=0; --t) {
    int n = t / K;
    int m = (t+1) / K;
    int o = (t+2) / K;

    for (stp=0; stp<S; ++stp)
      for (st=0; st<S; ++st)
    
	/* Is it necessary to compute everything? */
	
	if ((cnt !=0 && st != stp) ||
	    logstran[st][stp] == -INFINITY ||
	    logr[t][st] == -INFINITY) {

	  logpstgstpV1THtp[st][stp] = -INFINITY;
	  
	  continue;
	} else {
	  
	  double logpstpstgV1t, avpstgHtpstpV1t;
	  
	  matrix mht, *mHtpgV1t;

	  matrix pmean = darray_getmat(pair_mean, 1, stp, t+1);
	  matrix p2mom = darray_getmat(pair_2mom, 2, stp, t+1);
	  
	  matrix gtp = matrix_submatrix(&pmean, H, 2*H-1, 0, 0);
	  matrix Gtp = matrix_submatrix(&p2mom, H, 2*H-1, H, 2*H-1);

	  /* covht = F(:,:,st,t) */
	  
	  matrix covht = darray_getmat(F, 2, st, n);

	  /* P = CovHtpHtpgV1t(:,:,stp,st) */

	  matrix P = darray_getmat(CovHtpHtpgV1t, 2, stp, st);

	  /* Q = CovHtpHtgV1t(:,:,stp,st) */

	  matrix Q = darray_getmat(CovHtpHtgV1t, 2, stp, st);

	  /* Xp = leftA(:,:,stp,st)' */

	  matrix Xp = darray_getmat(leftAp, 2, stp, st);

	  /* Do we need to (re)compute leftA? */
	  
	  if (m != n || m != o || t == T-2) {

	    /* Q = A(:,:,stp)*covht */

	    matrix_dgemm(NoTrans, NoTrans, 1.0, A[stp], &covht, 0.0, &Q);

	    /* P = A(:,:,stp)*covht*A(:,:,stp)' + gainH(:,:,stp, (t+1)/K) */

	    matrix_dgemm(NoTrans, Trans, 1.0, &covht, A[stp], 0.0, tmp[2]);
	    matrix_dgemm(NoTrans, NoTrans, 1.0, A[stp], tmp[2], 0.0, &P);
	    
	    *(P.data) += gainH[m][stp];

	    if (gainN) {
	      /* P = P + gainN(:,:,stp, (t+1)/K) */

	      double a = matrix_get(&P, E, E);
	      matrix_set(&P, E, E, a + gainN[n][stp]);
	    }

	    /* Xp = inv(P) * Q; P = LU(P) */
	    
	    matrix_memcpy(&Xp, &Q);
	    matrix_dgesv(&P, &Xp, perm);
	    
	    /* avtmp = -0.5*log(det(P))-0.5*H*log(2*pi) */
	    
	    avtmp = -0.5*matrix_lu_lndet(&P)-H*half_log2pi;
	    
	    /* P = inv(P) */
	    
	    matrix_dgetri(&P, perm);
	  }

	  /* mht = f(:,st,t) */
	  
	  mht = darray_getmat(f, 1, st, t);
	  
	  /* mHtpgV1t = A(:,:,stp)*mht */
	  
	  matrix_dgemm(NoTrans, NoTrans, 1.0, A[stp], &mht, 0.0, tmp[0]);
	  
	  mHtpgV1t = tmp[0];
	  	  	  
	  /* logpstpstgV1t = logstran(stp,st) + logr(st,t) */
	  
	  logpstpstgV1t = logr[t][st] + (cnt == 0 ? logstran[st][stp] : 0.0);
	    	    	    	      
	  /* tmp5 = gtp - mHtpgV1t */
	      
	  matrix_addn(NoTrans, NoTrans, 1.0, &gtp, -1.0, mHtpgV1t, tmp[5]);
	      
	  /* mHtV1T(:,stp,it,st) = mht + X*tmp5 */
	      
	  {
	    matrix x = darray_getmat(mHtV1T, 1, stp, st);
	    
	    matrix_memcpy(&x, &mht);
	    matrix_dgemm(Trans, NoTrans, 1.0, &Xp, tmp[5], 1.0, &x);
	  }
	      
	  /* tmp6 = Gtp*X' - Q */
	      
	  matrix_memcpy(tmp[6], &Q);
	  matrix_dgemm(NoTrans, NoTrans, 1.0, &Gtp, &Xp, -1.0, tmp[6]);

	  /* CovHtV1T(:,:,stp,st) = F(:,:,st,t) + X*tmp6 */
	  
	  {
	    matrix x = darray_getmat(CovHtV1T, 2, stp, st);
	    
	    matrix_memcpy(&x, &covht);
	    matrix_dgemm(Trans, NoTrans, 1.0, &Xp, tmp[6], 1.0, &x); 
	  }
	      
	  /* tmp7 = inv(CovHtpHtpV1t) * tmp5 */
	      
	  matrix_dgemm(NoTrans, NoTrans, 1.0, &P, tmp[5], 0.0, tmp[7]);

	  if (useKim) {
	    logpstgstpV1THtp[st][stp] = logpstpstgV1t;
	  } else {
	    /* avpstgHtpstpV1t = avtmp - 0.5*tmp5'*tmp7 */
	    
	    avpstgHtpstpV1t = avtmp - 0.5*matrix_ddot(tmp[5], tmp[7]);
	    
	    /* p(st | stp, v1:t, htp) ~ p(st, stp | v1:t) <...> */
	    
	    logpstgstpV1THtp[st][stp] = logpstpstgV1t + avpstgHtpstpV1t;
	  }  

	  /* CovHtHtpV1T(:,:,stp,st) = X*G(:,:,stp,t+1) */
	      
	  {
	    matrix x = darray_getmat(CovHtHtpV1T, 2, stp, st);
	    
	    matrix_dgemm(Trans, NoTrans, 1.0, &Xp, &Gtp, 0.0, &x);
	  }
	}

    /* Normalise p(st | stp, v1:t, htp) */

    for (stp=0; stp<S; ++stp) {
      double logpstpgV1THtp = -INFINITY;
      
      for (st=0; st<S; ++st)
	logpstpgV1THtp = logadd(logpstpgV1THtp, logpstgstpV1THtp[st][stp]);
      
      if (logpstpgV1THtp != -INFINITY)
	for (st=0; st<S; ++st)
	  logpstgstpV1THtp[st][stp] -= logpstpgV1THtp;
    }
    
    /* p(stp, st | v1:T) = p(st | stp, v1:t, htp) * p(stp | v1:T) */
    
    {
      double logs = -INFINITY;
      
      for (st=0; st<S; ++st)
	for (stp=0; stp<S; ++stp) {
	  double tmp =  logpstgstpV1THtp[st][stp] + logl[t+1][stp];
	
	  logpstpstgV1T[st][stp] = tmp;
	
	  logs = logadd(logs, tmp);
	}

      for (st=0; st<S; ++st)
	for (stp=0; stp<S; ++stp)
	  logpstpstgV1T[st][stp] -= logs;
    }

    /* Collapse to a single Gaussians */

    for (st=0; st<S; ++st) {
      double logpstgV1T = logsum(S, logpstpstgV1T[st]);
      
      logl[t][st] = logpstgV1T;
      
      if (logpstgV1T != -INFINITY) {

	matrix pmean = darray_getmat(pair_mean, 1, st, t);
	matrix p2mom = darray_getmat(pair_2mom, 2, st, t);
	
	matrix gt = matrix_submatrix(&pmean, H, 2*H-1, 0, 0);
	matrix Gt = matrix_submatrix(&p2mom, H, 2*H-1, H, 2*H-1);

	darray means = darray_subarray(mHtV1T, 2, st);
	darray Covs  = darray_subarray(CovHtV1T, 3, st);

	arslds_mix2sgl(S, logpstpstgV1T[st], logpstgV1T, 
		       &means, &Covs, &gt, &Gt);
      }
    }

    /* Find pair means and second moment */

    for (stp=0; stp<S; ++stp) {
      double logpstpgV1T = logl[t+1][stp];

      if (logpstpgV1T != -INFINITY) {
	matrix pmean = darray_getmat(pair_mean, 1, stp, t+1);
	matrix p2mom = darray_getmat(pair_2mom, 2, stp, t+1);
	
	matrix sub1  = matrix_submatrix(&pmean, 0, H-1, 0, 0);
	matrix sub2  = matrix_submatrix(&pmean, H, 2*H-1, 0, 0);
	matrix sub11 = matrix_submatrix(&p2mom, 0, H-1, 0, H-1);
	matrix sub12 = matrix_submatrix(&p2mom, 0, H-1, H, 2*H-1);
	matrix sub21 = matrix_submatrix(&p2mom, H, 2*H-1, 0, H-1);
	matrix sub22 = matrix_submatrix(&p2mom, H, 2*H-1, H, 2*H-1);

	for (st=0; st<S; ++st) {
	  double pstgstpV1T = exp(logpstpstgV1T[st][stp] - logpstpgV1T);
	  
	  if (pstgstpV1T != 0) {
	    /* mht = mhtV1T(:,stp,st) */
	    
	    matrix mht = darray_getmat(mHtV1T, 1, stp, st);
	    
	    /* covht = CovHtV1T(:,:,stp,st) */
	    
	    matrix covht  = darray_getmat(CovHtV1T, 2, stp, st);
	    
	    /* covhthtp = CovHtHtpV1T(:,:,stp,st) */
	    
	    matrix covhthtp = darray_getmat(CovHtHtpV1T, 2, stp, st);

	    /* sub1 = sub1 + pstgstpV1T * mht */

	    matrix_addn(NoTrans, NoTrans, 1.0, &sub1, pstgstpV1T, &mht, &sub1); 
	    	    
	    /* sub11 = sub11 + pstgstpV1T * (covht + mht*mht') */

	    matrix_memcpy(tmp[8], &covht);
	    matrix_dgemm(NoTrans, Trans, 1.0, &mht, &mht, 1.0, tmp[8]);
	    matrix_addn(NoTrans, NoTrans, 1.0, &sub11, 
			pstgstpV1T, tmp[8], &sub11);
	    
	    /* sub12 = sub12 + pstgstpV1T * (covhthtp + mht*sub2') */

	    matrix_memcpy(tmp[8], &covhthtp);
	    matrix_dgemm(NoTrans, Trans, 1.0, &mht, &sub2, 1.0, tmp[8]);
	    matrix_addn(NoTrans, NoTrans, 1.0, &sub12,
			pstgstpV1T, tmp[8], &sub12);
	  }
	}

	/* sub21 = sub12' */
	
	matrix_memcpy(&sub21, &sub12);
	matrix_transpose(&sub21);

	/* sub22 = sub22 + sub2*sub2' */

	matrix_dgemm(NoTrans, Trans, 1.0, &sub2, &sub2, 1.0, &sub22);
      }
    }
    
    cnt = (cnt + 1) % K;
  }

  /* For each stp, convert G(:,:,stp,1) to 2mom. */

  for (stp=0; stp<S; ++stp)
    if (logl[0][stp] != -INFINITY) {
      matrix pmean = darray_getmat(pair_mean, 1, stp, 0);
      matrix p2mom = darray_getmat(pair_2mom, 2, stp, 0);

      matrix sub2  = matrix_submatrix(&pmean, H, 2*H-1, 0, 0);
      matrix sub22 = matrix_submatrix(&p2mom, H, 2*H-1, H, 2*H-1);
      
      /* sub22 = sub22 + sub2*sub2' */
      
      matrix_dgemm(NoTrans, Trans, 1.0, &sub2, &sub2, 1.0, &sub22);
    }
  
  /* Deallocate temporary variables */

  for (i=0; i<9; ++i)
    matrix_free(tmp[i]);
  
  /* Deallocate memory */
  
  for (i=0; i<S; ++i) {
    free(logstran[i]);
    free(logpstgstpV1THtp[i]);
    free(logpstpstgV1T[i]);
  }
  
  free(logstran);
  free(logpstgstpV1THtp);
  free(logpstpstgV1T);

  free(perm);
  
  darray_free(mHtV1T);
  darray_free(CovHtV1T);
  darray_free(CovHtHtpV1T);

  darray_free(CovHtpHtgV1t);
  darray_free(CovHtpHtpgV1t);
  darray_free(leftAp);
}
/*-------------------------------------------------------------------------*/
void arslds_mix2sgl(int N, double *logp, double logz,
		    darray *means, darray *Covs,
		    matrix *newmean, matrix *newCovMom) {

  int i;
  int H = newmean->nr;
  
  matrix *tmp0 = matrix_alloc(H,H);
  
  for (i=0; i<N; ++i) {
    double p = exp(logp[i] - logz);
    
    matrix mean  = darray_getmat(means, 1, i); 
    matrix sigma = darray_getmat(Covs, 2, i);
    
    /* newmean = newmean + p * mean */
    
    matrix_addn(NoTrans, NoTrans, 1.0, newmean, p, &mean, newmean);
    
    /* newCovMom = newCovMom + p * (sigma + mean * mean') */
    
    matrix_dgemm(NoTrans, Trans, 1.0, &mean, &mean, 0.0, tmp0);
    matrix_add(&sigma, tmp0, tmp0);
    matrix_addn(NoTrans, NoTrans, 1.0, newCovMom, p, tmp0, newCovMom);
  }
  
  /* newCovMom = newCovMom - newmean * newmean' */
  
  matrix_dgemm(NoTrans, Trans, -1.0, newmean, newmean, 1.0, newCovMom);

  /* newCovMom = 0.5*(newCovMom + newCovMom') */
  
  matrix_addm(NoTrans, Trans, newCovMom, newCovMom, tmp0);
  matrix_scale(0.5, tmp0);
  
  matrix_memcpy(newCovMom, tmp0);
  
  /* Deallocate memory */
  
  matrix_free(tmp0);
}
