/* arslds_fwd.c
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 */

#include <assert.h>
#include <string.h>

#include "mex.h"
#include "arslds.h"
#include "matrix.h"
#include "darray.h"

void matrix_mexprint(matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (i=0; i<nrm; ++i) {
    for (j=0; j<ncm; ++j)
      mexPrintf("%e ", *(dm + j*ldm + i));

    mexPrintf("\n");
  }
}

void arslds_mexload(arslds_t *arslds, mxArray const *data) {

  int i, j, size;
  int nd, nr, nc;
  int const *dims;
  double *x;

  int S, H, K, D;

  mxArray *ptr;

  /* Discrete prior */

  ptr  = mxGetField(data, 0, "prior");
  dims = mxGetDimensions(ptr);

  S = dims[0];
  x = mxGetPr(ptr);

  arslds->prior = (double*)malloc(dims[0]*sizeof(double));
  memcpy(arslds->prior, x, dims[0]*sizeof(double));

  /* Discrete transition matrix */

  ptr  = mxGetField(data, 0, "stran");
  dims = mxGetDimensions(ptr);
  nr   = dims[0];
  nc   = dims[1];

  x = mxGetPr(ptr);

  arslds->stran = (double**)malloc(nr*sizeof(double*));

  for (i=0; i<nr; ++i) {
    arslds->stran[i] = (double*)malloc(nc*sizeof(double));

    for (j=0; j<nc; ++j)
      arslds->stran[i][j] = x[i*nr+j];
  }

  /* Prior mean */

  ptr  = mxGetField(data, 0, "pMean");
  dims = mxGetDimensions(ptr);
  
  H = dims[0];
  x = mxGetPr(ptr);
  
  arslds->pMean = (matrix**)malloc(S*sizeof(matrix*));

  for (i=0; i<S; ++i) {
    arslds->pMean[i] = matrix_alloc(H,1);
    memcpy(arslds->pMean[i]->data, x+(i*H), H*sizeof(double));
  }

  /* Prior cov */

  ptr  = mxGetField(data, 0, "pCov");
  dims = mxGetDimensions(ptr);
  size = H*H;

  x = mxGetPr(ptr);

  arslds->pCov = (matrix**)malloc(S*sizeof(matrix*));

  for (i=0; i<S; ++i) {
    arslds->pCov[i] = matrix_alloc(H,H);
    memcpy(arslds->pCov[i]->data, x+(i*size), size*sizeof(double));
  }

  /* Hidden continuous transition matrix */

  ptr  = mxGetField(data, 0, "A");
  dims = mxGetDimensions(ptr);
  size = H*H;

  x = mxGetPr(ptr);

  arslds->A = (matrix**)malloc(S*sizeof(matrix*));

  for (i=0; i<S; ++i) {
    arslds->A[i] = matrix_alloc(H,H);
    memcpy(arslds->A[i]->data, x+(i*size), size*sizeof(double));
  }

  /* Hidden covariance matrix */

  ptr  = mxGetField(data, 0, "CovH");
  dims = mxGetDimensions(ptr);
  size = H*H;

  x = mxGetPr(ptr);

  arslds->CovH = (double*)malloc(S*sizeof(double));

  for (i=0; i<S; ++i)
    arslds->CovH[i] = x[i];
  
  /* Visible covariance matrix */

  ptr  = mxGetField(data, 0, "CovV");
  dims = mxGetDimensions(ptr);

  x = mxGetPr(ptr);

  arslds->CovV = (double*)malloc(S*sizeof(double));

  for (i=0; i<S; ++i)
    arslds->CovV[i] = x[i];

  /* Visible bias */

  ptr  = mxGetField(data, 0, "vbias");
  dims = mxGetDimensions(ptr);

  x = mxGetPr(ptr);

  arslds->vbias = (double*)malloc(S*sizeof(double));

  for (i=0; i<S; ++i)
    arslds->vbias[i] = x[i];

  /* State duration */

  ptr = mxGetField(data, 0, "K");
  
  if (ptr == 0) {
    K = 1;
  } else {
    x = mxGetPr(ptr);
    K = (int)x[0];
  }

  /* Order of the noise model */

  ptr = mxGetField(data, 0, "D");

  if (ptr == 0) {
    D = 0;
  } else {
    x = mxGetPr(ptr);
    D = (int)x[0];
  }

  /* ... */

  arslds->S = S;
  arslds->H = H;
  arslds->K = K;
  arslds->D = D;
}

void free_arslds(arslds_t *arslds) {
  int i;
  int S = arslds->S;

  for (i=0; i<S; ++i) {
    free(arslds->stran[i]);
    
    matrix_free(arslds->pMean[i]);
    matrix_free(arslds->pCov[i]);
    matrix_free(arslds->A[i]);
  }

  free(arslds->prior);
  free(arslds->stran);
  free(arslds->pMean);
  free(arslds->pCov);
  free(arslds->A);
  free(arslds->CovH);
  free(arslds->CovV);
  free(arslds->vbias);
}

darray *darray_mexalloc(mxArray const *data) {
  int i;
  int const *dims;
  darray *a = (darray*)malloc(sizeof(darray));

  int nd = mxGetNumberOfDimensions(data);
  
  a->nd  = nd;
  a->dim = (size_t*)malloc(nd * sizeof(size_t));
  a->dd  = (size_t*)malloc(nd * sizeof(size_t));
  a->own = 0;

  dims = mxGetDimensions(data);

  a->dim[0] = dims[0];
  a->dd[0]  = 1;

  for (i=1; i<nd; ++i) {
    a->dim[i] = dims[i];
    a->dd[i]  = a->dd[i-1] * a->dim[i-1];
  }

  a->ne   = a->dd[nd-1]*a->dim[nd-1];
  a->data = mxGetPr(data);

  return a;
}

void mexFunction(int nlhs, mxArray *plhs[],
		 int nrhs, mxArray const *prhs[]) {

  int i, j, t, size;
  double *x;
  int const *dims;
  int S, H, T, N;
  int tmp[5];

  double *v_samp, **gainH, **gainV, **gainN;

  arslds_t arslds;

  mxArray const *ptr;

  double llh;
  double **logr;
  darray *f, *F;

  /* Check the number of input/output arguments */

  if (nrhs < 4 || nrhs > 5) {
    mexErrMsgTxt("Not the right number of input arguments.");
  }

  if (nlhs != 4) {
    mexErrMsgTxt("Not the right number of output arguments.");
  }  

  /* First argument (the arslds) */

  arslds_mexload(&arslds, prhs[0]);

  S = arslds.S;
  H = arslds.H;

  /* Second argument (the visible sequence) */

  ptr  = prhs[1];
  dims = mxGetDimensions(ptr);
  T    = dims[1];

  x = mxGetPr(ptr);

  v_samp = (double*)malloc(T*sizeof(double));

  for (i=0; i<T; ++i)
    v_samp[i] = x[i];

  /* Third argument (gainH) */

  ptr  = prhs[2];
  dims = mxGetDimensions(ptr);
  N    = dims[1];

  x = mxGetPr(ptr);

  gainH = (double**)malloc(N*sizeof(double*));

  for (i=0; i<N; ++i) {
    gainH[i] = (double*)malloc(S*sizeof(double));
    
    for (j=0; j<S; ++j)
      gainH[i][j] = x[i*S+j];
  }

  /* Fourth argument (gainV) */

  ptr = prhs[3];
  x   = mxGetPr(ptr);

  gainV = (double**)malloc(N*sizeof(double*));

  for (i=0; i<N; ++i) {
    gainV[i] = (double*)malloc(S*sizeof(double));

    for (j=0; j<S; ++j)
      gainV[i][j] = x[i*S+j];
  }

  /* Fifth (optional) argument (gainN) */

  if (nrhs == 5) {
    ptr = prhs[4];
    x   = mxGetPr(ptr);

    gainN = (double**)malloc(N*sizeof(double*));

    for (i=0; i<N; ++i) {
      gainN[i] = (double*)malloc(S*sizeof(double));
      
      for (j=0; j<S; ++j)
	gainN[i][j] = x[i*S+j];
    }
  } else {
    gainN = 0;
  } 

  /* Output arguments */

  tmp[0] = S;
  tmp[1] = T;

  plhs[0] = mxCreateNumericArray(2, tmp, mxDOUBLE_CLASS, mxREAL);

  tmp[0] = H;
  tmp[1] = S;
  tmp[2] = T;

  plhs[1] = mxCreateNumericArray(3, tmp, mxDOUBLE_CLASS, mxREAL);

  tmp[0] = H;
  tmp[1] = H;
  tmp[2] = S;
  tmp[3] = N;

  plhs[2] = mxCreateNumericArray(4, tmp, mxDOUBLE_CLASS, mxREAL);

  plhs[3] = mxCreateDoubleMatrix(1, 1, mxREAL);

  /* ... */

  logr = (double**)malloc(T*sizeof(double*));

  for (t=0; t<T; ++t)
    logr[t] = (double*)malloc(S*sizeof(double));

  f = darray_mexalloc(plhs[1]);
  F = darray_mexalloc(plhs[2]);

  arslds_fwd_sgl(&arslds, T, v_samp, gainH, gainV, gainN, logr, f, F, &llh);

  /* Convert output arguments */

  /* Filtered posteriors */

  x = mxGetPr(plhs[0]);

  for (t=0; t<T; ++t) {
    memcpy(x, logr[t], S*sizeof(double));
    x += S;
  }

  /* Log likelihood */

  x  = mxGetPr(plhs[3]);
  *x = llh;

  /* Deallocate memory */

  free_arslds(&arslds);

  free(v_samp);

  for (t=0; t<T; ++t)
    free(logr[t]);

  free(logr);

  for (i=0; i<N; ++i) {
    free(gainH[i]);
    free(gainV[i]);
  }

  free(gainH);
  free(gainV);

  if (gainN) {
    for (i=0; i<N; ++i)
      free(gainN[i]);

    free(gainN);
  }

  darray_free(f);
  darray_free(F);
}
