/* linalg_matlab.c
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "linalg.h"
#include "matlab.h"
/*-------------------------------------------------------------------------*/
void matrix_dgemm(trans tra, trans trb, double alpha, 
		  matrix *a, matrix *b, double beta, matrix *c) {

  char *tra_ = tra == NoTrans ? "N" : "T";
  char *trb_ = trb == NoTrans ? "N" : "T";

  assert((tra == NoTrans ? a->nc : a->nr) == 
	 (trb == NoTrans ? b->nr : b->nc));
  assert(c->nr == (tra == NoTrans ? a->nr : a->nc));
  assert(c->nc == (trb == NoTrans ? b->nc : b->nr));

  dgemm_(tra_, trb_,
	 tra == NoTrans ? &a->nr : &a->nc, 
	 trb == NoTrans ? &b->nc : &b->nr, 
	 tra == NoTrans ? &a->nc : &a->nr, 
	 &alpha, a->data, &a->ld, b->data, &b->ld,
	 &beta, c->data, &c->ld);
} 
/*-------------------------------------------------------------------------*/
void matrix_dgemv(trans tra, double alpha, matrix *a,
		  matrix *x, double beta, matrix *y) {

  char *tra_ = tra == NoTrans ? "N" : "T";
  int dummy  = 1;

  dgemv_(tra,
	 &a->nr, &a->nc, 
	 &alpha, a->data, &a->ld, x->data, &dummy,
	 &beta, y->data, &dummy);
}
/*-------------------------------------------------------------------------*/
void matrix_dtrmv(trans tra, matrix *a, matrix *x) {

  char *tra_  = tra == NoTrans ? "N" : "T";
  char *uplo_ = "U";
  char *diag_ = "N";
  int  dummy  = 1;

  assert(a->nr == a->nc);

  dtrmv_(uplo_, tra_, diag_,
	 &a->nr, a->data, &a->ld, x->data, &dummy);
}
/*-------------------------------------------------------------------------*/
void matrix_dtrsv(trans tra, matrix *a, matrix *x) {

  char *tra_  = tra == NoTrans ? "N" : "T";
  char *uplo_ = "U";
  char *diag_ = "N";
  int  dummy  = 1;
  
  assert(a->nr == a->nc);

  dtrsv_(uplo_, tra_, diag_,
	 &a->nr, a->data, &a->ld, x->data, &dummy);
}
/*-------------------------------------------------------------------------*/
void matrix_daxpy(double alpha, matrix *x, matrix *y) {

  int dummy = 1;
  
  daxpy_(&x->nr, &alpha, x->data, &dummy, y->data, &dummy);
}
/*-------------------------------------------------------------------------*/
double matrix_ddot(matrix *x, matrix *y) {

  int dummy = 1;

  return ddot_(&x->nr, x->data, &dummy, y->data, &dummy);
}
/*-------------------------------------------------------------------------*/
void matrix_add(matrix *a, matrix *b, matrix *c) {

  int i, j;

  int nra = a->nr;
  int nca = a->nc;
  int lda = a->ld;
  int ldb = b->ld;
  int ldc = c->ld;

  double *da = a->data;
  double *db = b->data;
  double *dc = c->data;

  assert(a->nr == b->nr && a->nc == b->nc &&
	 a->nr == c->nr && a->nc == c->nc);

  for (j=0; j<nca; ++j) {
    double *ea = da + j*lda;
    double *eb = db + j*ldb;
    double *ec = dc + j*ldc;
    
    for (i=0; i<nra; ++i)
      *(ec + i) = *(ea + i) + *(eb + i);
  }
}
/*-------------------------------------------------------------------------*/
void matrix_addm(trans tra, trans trb, matrix *a, matrix *b, matrix *c) {

  int i, j;

  int nrc = c->nr;
  int ncc = c->nc;
  int lda = a->ld;
  int ldb = b->ld;
  int ldc = c->ld;

  double *da = a->data;
  double *db = b->data;
  double *dc = c->data;

  assert((tra == NoTrans ? a->nr : a->nc) == 
	 (trb == NoTrans ? b->nr : b->nc) && 
	 (tra == NoTrans ? a->nc : a->nr) ==
	 (trb == NoTrans ? b->nc : b->nr));
  
  assert(nrc == (tra == NoTrans ? a->nr : a->nc) &&
	 ncc == (tra == NoTrans ? a->nc : a->nr));
  
  for (j=0; j<ncc; ++j)
    for (i=0; i<nrc; ++i) {
      double *ea = da + (tra == NoTrans ? j : i) * lda;
      double *eb = db + (trb == NoTrans ? j : i) * ldb;
      double *ec = dc + j * ldc;
      
      int fa = tra == NoTrans ? i : j;
      int fb = trb == NoTrans ? i : j;

      *(ec + i) = *(ea + fa) + *(eb + fb);
    }
}
/*-------------------------------------------------------------------------*/
void matrix_addn(trans tra, trans trb, double alpha, matrix *a,
		 double beta, matrix *b, matrix *c) {

  int i, j;

  int nrc = c->nr;
  int ncc = c->nc;
  int lda = a->ld;
  int ldb = b->ld;
  int ldc = c->ld;

  double *da = a->data;
  double *db = b->data;
  double *dc = c->data;

  assert((tra == NoTrans ? a->nr : a->nc) == 
	 (trb == NoTrans ? b->nr : b->nc) && 
	 (tra == NoTrans ? a->nc : a->nr) ==
	 (trb == NoTrans ? b->nc : b->nr));
  
  assert(nrc == (tra == NoTrans ? a->nr : a->nc) &&
	 ncc == (tra == NoTrans ? a->nc : a->nr));

  for (j=0; j<ncc; ++j)
    for (i=0; i<nrc; ++i) {
      double *ea = da + (tra == NoTrans ? j : i) * lda;
      double *eb = db + (trb == NoTrans ? j : i) * ldb;
      double *ec = dc + j * ldc;
      
      int fa = tra == NoTrans ? i : j;
      int fb = trb == NoTrans ? i : j;

      *(ec + i) = alpha * *(ea + fa) + beta * *(eb + fb);
    }
}
/*-------------------------------------------------------------------------*/
void matrix_dgetrf(matrix *a, int *p) {

  int info;

  dgetrf_(&a->nr, &a->nc, a->data, &a->ld, p, &info);
}
/*-------------------------------------------------------------------------*/
void matrix_dgetri(matrix *a, int *p) {

  int lwork = -1;
  int info;

  double tmp;
  double *work;

  /* Get the optimal size of the work array */

  dgetri_(&a->nc, a->data, &a->ld, p, &tmp, &lwork, &info);

  lwork = (int)tmp;

  /* ... */

  work = (double*)malloc(lwork*sizeof(double));

  dgetri_(&a->nc, a->data, &a->ld, p, work, &lwork, &info);

  free(work);
}
/*-------------------------------------------------------------------------*/
void matrix_dgesv(matrix *a, matrix *b, int *p) {

  int info;

  dgesv_(&a->nc, &b->nc, a->data, &a->ld, p, b->data, &b->ld, &info);
}
/*-------------------------------------------------------------------------*/
double matrix_lu_det(matrix *a, int *p) {

  int i;
  double det = 1.0;

  int nra = a->nr;
  int lda = a->ld;

  double *da = a->data;

  assert(a->nr == a->nc);

  for (i=0; i<nra; ++i) {
    det *= *(da + i*lda + i);
    
    if (p[i] != i)
      det *= -1.0;
  }

  return det;
}
/*-------------------------------------------------------------------------*/
double matrix_lu_lndet(matrix *a) {

  int i;
  double det = 0.0;

  int nra = a->nr;
  int lda = a->ld;

  double *da = a->data;

  assert(a->nr == a->nc);

  for (i=0; i<nra; ++i)
    det += log(fabs( *(da + i*lda + i) ));

  return det;
}
/*-------------------------------------------------------------------------*/
void matrix_dpotrf(matrix *a) {

  char *uplo_ = "U";

  assert(a->nr == a->nc);
  
  dpotrf_(&uplo_, &a->nc, a->data, &a->ld);
}
/*-------------------------------------------------------------------------*/
double matrix_chol_sqrtdet(matrix *a) {
  
  int i;
  double det = 1.0;

  int nra = a->nr;
  int lda = a->ld;

  double *da = a->data;

  assert(a->nr == a->nc);

  for (i=0; i<nra; ++i)
    det *= *(da + i*lda + i);

  return det;
}
/*-------------------------------------------------------------------------*/
double matrix_chol_lnsqrtdet(matrix *a) {
  
  int i;
  double det = 0.0;

  int nra = a->nr;
  int lda = a->ld;

  double *da = a->data;

  assert(a->nr == a->nc);

  for (i=0; i<nra; ++i)
    det += log(fabs( *(da + i*lda + i) ));

  return det;
}
