/* matrix.c
 *
 * Copyright (C) 2006 Bertrand Mesot
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA 02110-1301, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "matrix.h"
/*-------------------------------------------------------------------------*/
matrix *matrix_alloc(int nr, int nc) {
  matrix *m = (matrix*)malloc(sizeof(matrix));
  
  m->data = (double*)malloc(nr*nc * sizeof(double));
  m->own  = 1;

  m->nr = nr;
  m->nc = nc;
  m->ld = nr;

  return m;
}
/*-------------------------------------------------------------------------*/
matrix *matrix_calloc(int nr, int nc) {
  matrix *m = (matrix*)malloc(sizeof(matrix));
  
  m->data = (double*)calloc(nr*nc, sizeof(double));
  m->own  = 1;

  m->nr = nr;
  m->nc = nc;
  m->ld = nr;

  return m;
}
/*-------------------------------------------------------------------------*/
void matrix_free(matrix *m) {
  if (m->own) 
    free(m->data);

  free(m);
}
/*-------------------------------------------------------------------------*/
matrix matrix_submatrix(matrix *m, int fr, int lr, int fc, int lc) {
  matrix mat;

  int nr = lr - fr + 1;
  int nc = lc - fc + 1;
  int ld = m->ld;

  assert(nr > 0 && nr <= m->nr);
  assert(nc > 0 && nc <= m->nc);

  mat.data = m->data + fc*ld + fr;
  mat.own  = 0;

  mat.nr = nr;
  mat.nc = nc;
  mat.ld = m->ld;

  return mat;
}
/*-------------------------------------------------------------------------*/
matrix matrix_diagonal(matrix *m) {
  matrix mat;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  mat.data = m->data;
  mat.own  = 0;

  mat.nr = 1;
  mat.nc = nrm < ncm ? nrm : ncm;
  mat.ld = ldm + 1;

  return mat;
}
/*-------------------------------------------------------------------------*/
void matrix_memcpy(matrix *dest, matrix *src) {
  int i;

  int nrs = src->nr;  
  int ncs = src->nc;
  int lds = src->ld;

  double *ds = src->data; 

  int nrd = dest->nr;
  int ncd = dest->nc;
  int ldd = dest->ld;

  double *dd = dest->data;

  assert(nrs == nrd && ncs == ncd);

  for (i=0; i<ncs; ++i) {
    double *es = ds + i*lds;
    double *ed = dd + i*ldd;

    memcpy(ed, es, nrs*sizeof(double));
  }
}
/*-------------------------------------------------------------------------*/
matrix *matrix_cpy(matrix *m) {
  int nr = m->nr;
  int nc = m->nc;

  matrix *mat = matrix_alloc(nr, nc);
  matrix_memcpy(mat, m);

  return mat;
}
/*-------------------------------------------------------------------------*/
void matrix_set(matrix *m, int i, int j, double v) {
  int ldm    = m->ld;
  double *dm = m->data;

  *(dm + j*ldm + i) = v;
}
/*-------------------------------------------------------------------------*/
double matrix_get(matrix *m, int i, int j) {
  int ldm    = m->ld;
  double *dm = m->data;

  return *(dm + j*ldm + i);
}
/*-------------------------------------------------------------------------*/
void matrix_set_identity(matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (j=0; j<ncm; ++j) {
    double *em = dm + j*ldm;
    
    for (i=0; i<nrm; ++i)
      *(em + i) = i == j ? 1.0 : 0.0;
  }
}
/*-------------------------------------------------------------------------*/
void matrix_init(matrix *m, double x) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (j=0; j<ncm; ++j) {
    double *em = dm + j*ldm;

    for (i=0; i<nrm; ++i)
      *(em + i) = x;
  }
}
/*-------------------------------------------------------------------------*/
void matrix_scale(double x, matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (j=0; j<ncm; ++j) {
    double *em = dm + j*ldm;

    for (i=0; i<nrm; ++i)
      *(em + i) *= x;
  }
}
/*-------------------------------------------------------------------------*/
void matrix_transpose(matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  assert(nrm == ncm);

  double *dm = m->data;

  for (j=0; j<ncm; ++j)
    for (i=j+1; i<nrm; ++i) {
      double *em1 = dm + j*ldm + i;
      double *em2 = dm + i*ldm + j;
      double tmp  = *em1;

      *em1 = *em2;
      *em2 = tmp;
    }
}
/*-------------------------------------------------------------------------*/
void matrix_randint(matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (j=0; j<ncm; ++j) {
    double *em = dm + j*ldm;

    for (i=0; i<nrm; ++i)
      *(em + i) = rand() % 100;
  }
}
/*-------------------------------------------------------------------------*/
void matrix_print(int trans, matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  if (trans)
    for (j=0; j<ncm; ++j) {
      for (i=0; i<nrm; ++i)
	printf("%e ", *(dm++));

      printf("\n");
    }
  else
    for (i=0; i<nrm; ++i) {
      for (j=0; j<ncm; ++j)
	printf("%e ", *(dm + j*ldm + i));
      
    printf("\n");
  }
}
/*-------------------------------------------------------------------------*/
void matrix_fprint(FILE *file, matrix *m) {
  int i, j;

  int nrm = m->nr;
  int ncm = m->nc;
  int ldm = m->ld;

  double *dm = m->data;

  for (i=0; i<nrm; ++i) {
    for (j=0; j<ncm; ++j)
      fprintf(file, "%e ", *(dm + j*ldm + i));

    fprintf(file, "\n");
  }
}
/*-------------------------------------------------------------------------*/
